#!/usr/bin/env python3

import json
import importlib.resources

import pandas as pd


# The maximum number of least frequent emojis to output in the table
EMOJI_DISPLAY_COUNT = 20


# Fetch a list of the most popular emoji on the Web to use as a basis for
# computing this metric
def get_emoji_list():

    return json.loads(importlib.resources.files(__package__)
                      .joinpath('emojis.json').read_text())


# Output the occurrences of specific emojis
def analyze(dfs):

    emojis = get_emoji_list()
    least_frequent_emojis = pd.DataFrame({
        'emoji': emojis,
        'count': [dfs.messages['text'].str.extract('(' + emoji + ')')
                  .count().item() for emoji in emojis]
    })
    # Filter out all emojis that have never been sent
    least_frequent_emojis = least_frequent_emojis[
        (least_frequent_emojis[['count']] != 0).all(axis=1)]
    return (least_frequent_emojis
            .sort_values(by='count', ascending=True)
            .reset_index(drop=True)
            .head(EMOJI_DISPLAY_COUNT))



#!/usr/bin/env python3


import pandas as pd


# The format to use for all date strings
DATE_FORMAT = '%Y-%m-%d'


def analyze(dfs):

    # Count all "text" column values by converting them to integers (always 1),
    # because resampling the DataFrame will remove all non-numeric columns
    dfs.messages['text'] = dfs.messages['text'].apply(
        pd.to_numeric, errors='coerce').isna()
    groups_by_day = dfs.messages.resample('D', on='datetime')
    sums_by_day = groups_by_day.sum()
    # Remove 00:00:00 from date index
    sums_by_day.index = sums_by_day.index.strftime(DATE_FORMAT)
    sums_by_day.index.rename('date', inplace=True)
    sums_by_day['is_from_them'] = (sums_by_day['text']
                                   - sums_by_day['is_from_me'])
    return sums_by_day



#!/usr/bin/env python3

import importlib.machinery
import importlib.util
import importlib.resources
import os
import os.path
import sqlite3
import sys
import types

import pandas as pd
from tabulate import tabulate


# In order to interpolate the user-specified list of chat identifiers into the
# SQL queries, we must join the list into a string delimited by a common
# symbol, then perform a LIKE comparison on this string within the SQL query;
# this is because named SQL parameters only support string values, rather than
# variable-length sequences
CHAT_IDENTIFIER_DELIMITER = '|'


# The path to the database file for the macOS Messages application
DB_PATH = os.path.expanduser(os.path.join(
    '~', 'Library', 'Messages', 'chat.db'))


# Join the list of chat identifiers into a delimited string; in order for the
# SQL comparison to function properly, this string must also start and end with
# the delimiter symbol
def get_chat_identifier_str(chat_identifiers):
    return '{start}{joined}{end}'.format(
        start=CHAT_IDENTIFIER_DELIMITER,
        joined=CHAT_IDENTIFIER_DELIMITER.join(chat_identifiers),
        end=CHAT_IDENTIFIER_DELIMITER)


# Return a pandas dataframe representing all messages in a particular
# conversation (identified by the given phone number)
def get_messages_dataframe(connection, chat_identifiers):

    return pd.read_sql_query(
        sql=importlib.resources.files(
            __package__).joinpath('queries/messages.sql').read_text(),
        con=connection,
        params={
            'chat_identifiers': get_chat_identifier_str(chat_identifiers),
            'chat_identifier_delimiter': CHAT_IDENTIFIER_DELIMITER
        },
        parse_dates={
            'datetime': {'infer_datetime_format': True}
        })


# Return a pandas dataframe representing all attachments in a particular
# conversation (identified by the given phone number)
def get_attachments_dataframe(connection, chat_identifiers):

    return pd.read_sql_query(
        sql=importlib.resources.files(
            __package__).joinpath('queries/attachments.sql').read_text(),
        con=connection,
        params={
            'chat_identifiers': get_chat_identifier_str(chat_identifiers),
            'chat_identifier_delimiter': CHAT_IDENTIFIER_DELIMITER
        })


# Return all dataframes for a specific macOS Messages conversation
def get_dataframes(chat_identifiers):

    with sqlite3.connect(DB_PATH) as connection:
        return types.SimpleNamespace(
            messages=get_messages_dataframe(
                connection, chat_identifiers),
            attachments=get_attachments_dataframe(
                connection, chat_identifiers))


# Load the given metric file as a Python module, and return the DataFrame
# provided by its analyze() function
def run_analyzer_for_metric_file(metric_file, dfs):
    loader = importlib.machinery.SourceFileLoader('metric_file', metric_file)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    module = importlib.util.module_from_spec(spec)
    # Expose package information to dynamically-imported module
    module.__package__ = __package__
    loader.exec_module(module)
    metric_df = module.analyze(dfs)
    return metric_df


# Format the given header name to be more human-readable (e.g. "foo_bar" =>
# "Foo Bar")
def prettify_header_name(header_name):
    if header_name:
        return header_name.replace('_', ' ').title()
    else:
        return None


# Format the given sequence of header names
def prettify_header_names(header_names):
    return [prettify_header_name(header_name) for header_name in header_names]


# Print the given dataframe of metrics data
def print_metrics(metric_df, format):
    # Prettify header row (i.e. column names)
    if metric_df.index.name:
        metric_df.index.rename(
            prettify_header_name(metric_df.index.name), inplace=True)
    metric_df.columns = prettify_header_names(metric_df.columns)
    # Prettify header column (i.e. textual values in first column)
    first_column_name = metric_df.columns[0]
    if metric_df[first_column_name].dtypes == object:
        metric_df[first_column_name] = metric_df[first_column_name].apply(
            prettify_header_name)

    # Make all indices start from 1 instead of 0, but only if the index is the
    # default (rather than a custom column)
    is_default_index = (not metric_df.index.name)
    if is_default_index:
        metric_df.index += 1

    # Output executed DataFrame to correct format
    if format == 'csv':
        print(metric_df.to_csv(
            index=not is_default_index,
            header=metric_df.columns))
    else:
        print(tabulate(metric_df,
                       showindex=not is_default_index,
                       headers=metric_df.columns))


# Analyze the macOS Messages conversation with the given recipient phone number
def analyze_conversation(chat_identifiers, metric_file, format):

    dfs = get_dataframes(chat_identifiers)

    # Quit if no messages were found for the specified conversation
    if not len(dfs.messages.index):
        print('No conversations found for the following handlers:\n{}'.format(
            '\n'.join(chat_identifiers)),
            file=sys.stderr)
        sys.exit(1)

    # Load and execute the given Python file as a module to retrieve the
    # relevant DataFrame
    metric_df = run_analyzer_for_metric_file(metric_file, dfs)

    # Print the given metrics
    print_metrics(metric_df, format)



#!/usr/bin/env python3

import datetime

import pandas as pd


# The format to use for all date strings
DATE_FORMAT = '%Y-%m-%d'


# Retrieve the date of the very first message sent in the conversation
def get_first_message_date(dfs):

    datestr = str(dfs.messages['datetime'].min().strftime(DATE_FORMAT))
    return datetime.datetime.strptime(datestr, DATE_FORMAT)


# Get all dates between (and including) the given two dates
def get_dates_between(start_date, end_date):

    return list(pd.date_range(start_date, end_date).strftime(DATE_FORMAT))


# Get all dates sent
def get_all_message_datestrs(dfs):

    groups_by_day = dfs.messages.resample('D', on='datetime')
    sums_by_day = groups_by_day.count()
    sums_by_day.index = sums_by_day.index.strftime(DATE_FORMAT)
    return list(sums_by_day.index)


# Get the number of messages with no reply (i.e. only one message sent for that
# day)
def get_noreply_count(dfs):

    # Count all "text" column values by converting them to integers (always 1),
    # because resampling the DataFrame will remove all non-numeric columns
    dfs.messages['text'] = dfs.messages['text'].apply(
        pd.to_numeric, errors='coerce').isna()
    groups_by_day = dfs.messages.resample('D', on='datetime')
    sums_by_day = groups_by_day.sum()
    sums_by_day['is_from_them'] = (sums_by_day['text']
                                   - sums_by_day['is_from_me'])
    return (sums_by_day['is_from_them'].eq(0).sum()
            + sums_by_day['is_from_me'].eq(0).sum())


# Return a list of all dates in Date List B which are not in Date List A
def get_missing_datestrs(date_list_a, date_list_b):

    return sorted(set(date_list_a) - set(date_list_b))


def analyze(dfs):

    all_datestrs = get_dates_between(
        get_first_message_date(dfs), str(datetime.date.today()))
    message_datestrs = get_all_message_datestrs(dfs)

    totals_map = {
        'messages': len(dfs.messages.index),
        'messages_from_me': dfs.messages['is_from_me'].eq(True).sum(),
        'messages_from_them': dfs.messages['is_from_me'].eq(False).sum(),
        'days_messaged': len(message_datestrs),
        'days_missed': len(all_datestrs) - len(message_datestrs),
        'days_with_no_reply': get_noreply_count(dfs),
    }
    return pd.DataFrame({
        'metric': tuple(totals_map.keys()),
        'total': tuple(totals_map.values())
    })



#!/usr/bin/env python3

import importlib.resources
import glob
import os
import os.path
import re
import sqlite3
import sys

import pandas as pd


# The glob pattern matching all AddressBook SQL databases to read from
DB_GLOB = os.path.expanduser(os.path.join(
    '~', 'Library', 'Application Support', 'AddressBook',
    'Sources', '*', '*.abcddb'))


# Normalize the given phone number to the format required for searching in the
# iMessage database
def normalize_phone_number(phone_number):

    # Strip all non-digits
    normalized_phone_number = re.sub('[^0-9]', '', phone_number)
    # Phone number must have area code
    if len(normalized_phone_number) < 10:
        raise RuntimeError(f'Phone number {phone_number} missing area code')
    # Add US country code if missing
    if len(normalized_phone_number) == 10:
        normalized_phone_number = '1' + normalized_phone_number
    normalized_phone_number = '+' + normalized_phone_number
    return normalized_phone_number


# Normalize the given email address to the format required for searching in the
# iMessage database
def normalize_email_address(email_address):

    normalized_email_address = email_address.strip()
    return normalized_email_address


# Fetch the sequence of chat identifiers for the contact with the given name
def get_chat_identifiers(contact_name):

    chat_identifiers = set()
    # There is a separate AddressBook database file for each "source" of
    # contacts (e.g. On My Mac, iCloud, etc.); we must query each of these
    # databases and combine the separate results into a single result set
    for db_path in glob.iglob(DB_GLOB):
        with sqlite3.connect(db_path) as connection:
            rows = pd.read_sql_query(
                sql=importlib.resources.files(
                    __package__).joinpath('queries/contact.sql').read_text(),
                con=connection,
                params={'contact_name': contact_name})
            # Combine the results
            chat_identifiers.update(
                normalize_phone_number(phone_number)
                for phone_number in rows['ZFULLNUMBER']
                if phone_number)
            chat_identifiers.update(
                normalize_email_address(email_address)
                for email_address in rows['ZADDRESS']
                if email_address)

    # Quit if the contact with the specified name could not be found
    if not len(chat_identifiers):
        print('No contact found with the name "{}"'.format(contact_name),
              file=sys.stderr)
        sys.exit(1)

    return sorted(chat_identifiers)



#!/usr/bin/env python3

import argparse
import os
import os.path

import ica.contact as contact
import ica.analyzer as analyzer


# Parse user arguments from the command line
def get_cli_args():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--contact-name',
        '-c',
        required=True)
    parser.add_argument(
        '--metric-file',
        '-m',
        type=os.path.expanduser,
        required=True)
    parser.add_argument(
        '--format',
        '-f',
        choices=('csv',))

    return parser.parse_args()


# Program entry point
def main():

    cli_args = get_cli_args()

    chat_identifiers = contact.get_chat_identifiers(
        contact_name=cli_args.contact_name)

    try:
        analyzer.analyze_conversation(
            chat_identifiers=chat_identifiers,
            metric_file=cli_args.metric_file,
            format=cli_args.format)
    except KeyboardInterrupt:
        print('')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3


import pandas as pd


def analyze(dfs):

    totals_map = {
        'gifs': dfs.attachments['mime_type'].eq('image/gif').sum(),
        'youtube_videos': (dfs.messages['text'].str.extract(
            r'(https?://(?:www\.)(?:youtube\.com|youtu\.be)/(?:.*?)(?:\s|$))')
            .drop_duplicates()
            .count()
            .item()),
        'apple_music': (dfs.messages['text'].str.extract(
            r'(https?://(?:music\.apple\.com)/(?:.*?)(?:\s|$))')
            .drop_duplicates()
            .count()
            .item()),
        'spotify': (dfs.messages['text'].str.extract(
            r'(https?://(?:open\.spotify\.com)/(?:.*?)(?:\s|$))')
            .drop_duplicates()
            .count()
            .item()),
        'recorded_videos': (dfs.attachments['mime_type']
                            .eq('video/quicktime').sum())
    }
    attachment_totals = pd.DataFrame({
        'type': totals_map.keys(),
        'total': totals_map.values()
    })
    return (attachment_totals
            .sort_values(by='total', ascending=False)
            .reset_index(drop=True))



#!/usr/bin/env python3

import pandas as pd


# Export the entire conversation
def analyze(dfs):

    return pd.DataFrame({
        'timestamp': dfs.messages['datetime'],
        # Convert 1/0 to Yes/No
        'is_from_me': (dfs.messages['is_from_me']
                       .apply(str)
                       .replace('1', 'Yes')
                       .replace('0', 'No')),
        # U+FFFC is the object replacement character, which appears as the
        # textual message for every attachment
        'message': dfs.messages['text'].replace(
            r'\ufffc', '(attachment)', regex=True)
    })
